from typing import (
    Any,
    List,
    Tuple,
    Union,
)

import sqlalchemy


class BaseField:
    def __init__(self, name: str = None, field: sqlalchemy.Column = None):
        self.name = name
        self.field = field

    def cast(self, value) -> Any:
        return value

    def run_validator(self, value: Any) -> Tuple[bool, List[str]]:
        errors = []
        for method_name in dir(self):
            if not method_name.startswith("check_if_"):
                continue

            executable = getattr(self, method_name)

            cast_value = self.cast(value)
            ok, error = executable(cast_value)
            if not ok:
                errors.append(error)

        return len(errors) <= 0, errors

    def check_if_required_and_filled(
        self, value
    ) -> Tuple[bool, Union[str, None]]:
        if not self.field.nullable and (value is None or value == ""):
            return False, "Can't be blank"

        return True, None


class BooleanField(BaseField):
    VALID_TRUES: List = ["True", "T", "true", "t", True, 1, "1"]

    VALID_FALSES: List = ["False", "F", "false", "f", False, 0, "0"]

    def cast(self, value) -> bool:
        if value in self.VALID_TRUES:
            return True
        elif value in self.VALID_FALSES:
            return False

        return value

    def check_if_is_a_valid_boolean(
        self, value
    ) -> Tuple[bool, Union[str, None]]:
        if not isinstance(value, bool):
            return False, "Not a valid boolean"

        return True, None


class IntegerField(BaseField):
    def cast(self, value) -> int:
        if value is None:
            return 0

        return int(value)


class StringField(BaseField):
    def cast(self, value) -> str:
        if value is None:
            return ""

        return str(value)

    def check_if_length_is_under_limit(
        self, value: str
    ) -> Tuple[bool, Union[str, None]]:
        if self.field.type.length and len(value) > self.field.type.length:
            return False, f"Limit of characters is {self.field.type.length}"

        return True, None


class MethodField(BaseField):
    pass



import os
from setuptools import setup


with open(os.path.join("README.md")) as file:
    readme = file.read()


with open(os.path.join("requirements.txt")) as file:
    requirements = file.read().split("\n")


setup(
    name="alcherializer",
    packages=["alcherializer"],
    version=os.getenv("GITHUB_REF").replace("refs/tags/v", ""),
    license="MIT",
    description="Django like model serializer",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Vinicius Guedes",
    author_email="viniciusgued@gmail.com",
    url="https://github.com/vinyguedess/alcherializer",
    download_url="https://github.com/vinyguedess/alcherializer/archive/master.zip",
    keywords=["django", "flask", "serializer", "sql", "sqlalchemy", "alchemy"],
    install_requires=requirements,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



from typing import (
    Any,
    Dict,
)

import sqlalchemy
from sqlalchemy.ext.declarative import DeclarativeMeta

from alcherializer import fields
from alcherializer.exceptions import MalformedMetaClassException


class Serializer:
    def __init__(self, instance=None, data=None, **kwargs):
        if not hasattr(self, "Meta") or not hasattr(self.Meta, "model"):
            raise MalformedMetaClassException("Serializer bad definition")

        self.meta = getattr(self, "Meta")

        self.errors = {}
        self.fields = self._get_fields()
        self.instance = instance
        self.initial_data = {} if not data else data
        self.many = kwargs.get("many", False)
        self.partial = kwargs.get("partial", False)
        self.validated_data = {}

    @property
    def data(self) -> Dict[str, Any]:
        instances = self.instance if self.many else [self.instance]

        results = []
        for instance in instances:
            obj_dict = {}
            for key in self.fields.keys():
                obj_dict[key] = self._get_instance_field_value(instance, key)

            results.append(obj_dict)

        return results if self.many else results[0]

    def clear(self) -> None:
        self.errors = {}
        self.validated_data = {}

    def is_valid(self) -> bool:
        for key, field in self.fields.items():
            if key in getattr(self.meta, "except_fields", ["id"]):
                continue

            value = self._get_value(key)
            valid, errors = field.get("validator").run_validator(value)
            if not valid:
                self.errors[key] = errors
                continue

            self.validated_data[key] = value

        return self._has_errors()

    def _get_value(self, key: str) -> Any:
        return self.initial_data.get(key)

    def _has_errors(self) -> bool:
        return len(self.errors.keys()) <= 0

    def _get_fields(self) -> Dict[str, Any]:
        required_fields = []
        if hasattr(self.meta, "fields"):
            required_fields = self.meta.fields

        columns = {}
        for key, value in self.meta.model.__dict__.items():
            if key.startswith("_"):
                continue

            if required_fields and key not in required_fields:
                continue

            columns[key] = {
                "type": value.type if hasattr(value, "type") else value,
                "required": value.nullable is False
                if hasattr(value, "nullable")
                else False,
                "validator": self._get_field_validator(key, value),
            }

        for field in required_fields:
            if field in columns:
                continue

            if not hasattr(self, field):
                continue

            columns[field] = {
                "type": getattr(self, field),
                "required": False,
                "validator": self._get_field_validator(field, field),
            }

        return columns

    def _get_field_validator(self, key: str, field):
        if hasattr(self, key):
            validator = getattr(self, key)
            validator.name = key
            validator.field = field

            return validator

        if isinstance(field.type, sqlalchemy.String):
            return fields.StringField(key, field)

        if isinstance(field.type, sqlalchemy.Integer):
            return fields.IntegerField(key, field)

        if isinstance(field.type, sqlalchemy.Boolean):
            return fields.BooleanField(key, field)

        return fields.BaseField(key, field)

    def _get_instance_field_value(self, instance, field: str) -> Any:
        if isinstance(self.fields[field].get("type"), fields.MethodField):
            return getattr(self, f"get_{field}", lambda o: None)(instance)

        value = getattr(instance, field)
        if isinstance(value.__class__, DeclarativeMeta):
            serializer = self.fields[field]["validator"]
            if isinstance(serializer, Serializer):
                serializer.instance = value
                return serializer.data

        if (
            isinstance(value, list)
            and len(value) > 0
            and isinstance(value[0].__class__, DeclarativeMeta)
        ):
            serializer = self.fields[field]["validator"]
            if isinstance(serializer, Serializer):
                serializer.instance = value
                serializer.many = True
                return serializer.data

        return value
